from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict, alias_generators

from yaml_ld.models import URI, JsonLdRecord

ExtractAllScripts = Annotated[
    bool,
    'Extract all JSON-LD script elements, as opposed to only the first one.',
]


class ExtractAllScriptsOptions(BaseModel):    # type: ignore
    """Options flag to extract all scripts or not."""

    extract_all_scripts: bool = False
    """
    True to extract all YAML-LD script elements from HTML, False to extract just
    the first.
    """


class ExpandContextOptions(BaseModel):    # type: ignore
    """Options flag to extract all scripts or not."""

    expand_context: JsonLdRecord | URI | None = None
    """A context to expand with."""

    model_config = ConfigDict(arbitrary_types_allowed=True)


def _default_document_loader():
    from yaml_ld.document_loaders.default import (  # noqa: WPS433
        DEFAULT_DOCUMENT_LOADER,
    )
    return DEFAULT_DOCUMENT_LOADER


class BaseOptions(BaseModel):   # type: ignore
    """Base options shared by all YAML-LD API methods."""

    base: URI | None = None
    """The base IRI to use."""

    document_loader: Any = None   # type: ignore
    """The document loader."""

    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=alias_generators.to_camel,
        arbitrary_types_allowed=True,
        validate_default=False,
    )



from abc import ABC, abstractmethod

from typing_extensions import TypedDict

from yaml_ld.models import URI, RemoteDocument

DocumentLoaderOptions = TypedDict(
    'DocumentLoaderOptions',
    {
        'extractAllScripts': bool,
        'headers': dict[str, str],
    },
)


class DocumentLoader(ABC):
    @abstractmethod
    def __call__(
        self,
        source: URI,
        options: DocumentLoaderOptions,
    ) -> RemoteDocument:
        """Load a document."""
        raise NotImplementedError()



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.content_types import DEFAULT_ACCEPT_HEADER
from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    JsonLdInput,
    ensure_string_or_document,
)
from yaml_ld.options import BaseOptions, ExtractAllScriptsOptions
from yaml_ld.rdf import Dataset


class ToRDFOptions(BaseOptions, ExtractAllScriptsOptions):   # type: ignore
    """Options for converting ＊-LD to RDF."""

    format: str | None = None
    """The format to use to output a string: 'application/n-quads'
    for N-Quads."""

    produce_generalized_rdf: bool = False
    """True to output generalized RDF, false to produce only standard RDF."""

    rdf_direction: str = 'i18n-datatype'
    """Only 'i18n-datatype' supported."""


DEFAULT_TO_RDF_OPTIONS = ToRDFOptions()


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def to_rdf(
    document: JsonLdInput,
    options: ToRDFOptions = DEFAULT_TO_RDF_OPTIONS,
) -> Dataset | str:
    """Convert a [＊-LD](/blog/any-ld/) document to RDF."""
    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)
    dict_options.setdefault('headers', {'Accept': DEFAULT_ACCEPT_HEADER})
    dict_options['extractAllScripts'] = True

    with except_json_ld_errors():
        jsonld._resolved_context_cache = jsonld.LRUCache(  # noqa: WPS437
            maxsize=jsonld.RESOLVED_CONTEXT_CACHE_MAX_SIZE,
        )
        return jsonld.to_rdf(
            input_=ensure_string_or_document(document),
            options=dict_options,
        )



import io

from rdflib import Graph
from rdflib_pyld_compat import pyld_jsonld_from_rdflib_graph

from yaml_ld.document_parsers.base import (
    BaseDocumentParser,
    DocumentLoaderOptions,
)
from yaml_ld.models import JsonLdRecord


class TurtleParser(BaseDocumentParser):
    """Parse Turtle documents."""

    def __call__(   # noqa: WPS238, WPS231, WPS225, C901
        self,
        data_stream: io.BytesIO,
        source: str,
        options: DocumentLoaderOptions,
    ) -> JsonLdRecord | list[JsonLdRecord]:
        """Parse YAML document stream into LD."""
        graph = Graph().parse(data_stream, format='turtle')
        return pyld_jsonld_from_rdflib_graph(graph)



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.content_types import DEFAULT_ACCEPT_HEADER
from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.models import DEFAULT_VALIDATE_CALL_CONFIG, RemoteDocument
from yaml_ld.options import BaseOptions

DEFAULT_BASE_OPTIONS = BaseOptions()


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def load_document(
    url,
    base=None,
    profile=None,
    requestProfile=None,
    options: BaseOptions = DEFAULT_BASE_OPTIONS,
) -> RemoteDocument:
    """
    Load an [＊-LD](blog/any-ld/) document.

    The document can be retrieved from local filesystem or from the Web.
    """
    headers = {
        'Accept': DEFAULT_ACCEPT_HEADER,
    }

    if requestProfile:
        headers['Accept'] = (
            f'application/ld+json;profile={requestProfile}, '
        ) + headers['Accept']

    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)
    dict_options.setdefault('headers', headers)

    return jsonld.load_document(
        url=str(url),
        options=dict_options,
        base=base,
        profile=profile,
        requestProfile=requestProfile,
    )



from pathlib import Path
from typing import Any, Sequence

from pydantic import ConfigDict
from typing_extensions import TypedDict
from yarl import URL

JsonLdRecord = dict[str, Any]  # type: ignore


class RemoteDocument(TypedDict, total=False):
    """Specification of the remote document."""

    contentType: str
    contextUrl: str | None
    document: JsonLdRecord | list[JsonLdRecord]
    documentUrl: str
    profile: str | None


URI = str | URL | Path
"""
A Universal Resource Identifier can be represented as a `Path` or as a `URL`.

Or, it can be a `str`, and we will try to automatically identify what that is.
"""


JsonLdInput = (
    JsonLdRecord | Sequence[JsonLdRecord] | URI
)
"""
Input for `expand()`, `compact()` and other functions.

[Specification](https://w3c.github.io/json-ld-api/#dom-jsonldrecord)
"""

JsonLdContext = JsonLdRecord | list[JsonLdRecord | URI] | URI
"""
The `JsonLdContext` interface is used to refer to a value that may be a
`JsonLdRecord`, a sequence of `JsonLdRecord`-s, or a string representing an IRI,
which can be dereferenced to retrieve a valid JSON document.

[Specification](https://w3c.github.io/json-ld-api/#dom-jsonldcontext)
"""

DEFAULT_VALIDATE_CALL_CONFIG = ConfigDict(
    arbitrary_types_allowed=True,
    validate_default=False,
)


class Undefined:
    """Undefined."""


def ensure_string_or_document(
    input_: JsonLdInput,
) -> JsonLdRecord | Sequence[JsonLdRecord] | str | RemoteDocument:
    """Prepare `input_` for `pyld` functions."""
    if isinstance(input_, (URL, Path)):
        return str(input_)

    return input_



from typing import Annotated

from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.content_types import DEFAULT_ACCEPT_HEADER
from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    JsonLdContext,
    JsonLdInput,
    JsonLdRecord,
    ensure_string_or_document,
)
from yaml_ld.options import (
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
)


class CompactOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options to compact a YAML-LD document."""

    compact_arrays: bool = True
    """Compact arrays to single values when appropriate?"""

    graph: bool = False
    """True to always output a top-level graph."""

    skip_expansion: bool = False
    """True to skip the expansion process, False to include it."""


DEFAULT_COMPACT_OPTIONS = CompactOptions()


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def compact(  # noqa: WPS211
    document: JsonLdInput,
    ctx: Annotated[JsonLdContext | None, 'Context to compact with.'] = None,
    options: CompactOptions = DEFAULT_COMPACT_OPTIONS,
) -> JsonLdRecord | list[JsonLdRecord]:
    """
    Compact a [＊-LD](/blog/any-ld/) document.

    Replace full IRIs with shorter terms and compact IRIs using a context,
    making the document more human-readable while preserving its original
    structure and semantics.
    """
    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)
    dict_options.setdefault('headers', {'Accept': DEFAULT_ACCEPT_HEADER})

    with except_json_ld_errors():
        return jsonld.compact(
            input_=ensure_string_or_document(document),
            ctx=ctx or {},
            options=dict_options,
        )



from pathlib import Path

from yarl import URL

from yaml_ld.document_loaders import content_types
from yaml_ld.document_loaders.base import DocumentLoader, DocumentLoaderOptions
from yaml_ld.errors import LoadingDocumentFailed, NotFound
from yaml_ld.models import URI, RemoteDocument


class LocalFileDocumentLoader(DocumentLoader):
    """Load documents from a local file system."""

    def __call__(
        self,
        source: URI,
        options: DocumentLoaderOptions,
    ) -> RemoteDocument:
        """Load documents from a local file system."""
        path = Path(URL(str(source)).path)

        content_type = content_types.by_extension(path.suffix)
        if content_type is None:
            raise ValueError(f'What content type is extension {path.suffix}?')

        parser = content_types.parser_by_content_type(
            content_type=content_type,
            uri=str(path),
        )
        if parser is None:
            raise LoadingDocumentFailed(path=path)

        try:
            with path.open(mode='rb') as data_stream:
                yaml_document = parser(
                    data_stream=data_stream,   # type: ignore
                    source=str(source),
                    options=options,
                )
        except FileNotFoundError as file_not_found:
            raise NotFound(path) from file_not_found

        return {
            'document': yaml_document,
            'documentUrl': str(source),
            'contextUrl': None,
            'contentType': content_type,
        }



import contextlib
from json import JSONDecodeError

from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.content_types import DEFAULT_ACCEPT_HEADER
from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.errors import (
    CycleDetected,
    InvalidJSONLiteral,
    InvalidScriptElement,
    LoadingRemoteContextFailed,
    MappingKeyError,
    PyLDError,
)
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    JsonLdInput,
    JsonLdRecord,
    ensure_string_or_document,
)
from yaml_ld.options import (
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
)


class ExpandOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options for `jsonld.expand()`."""


DEFAULT_EXPAND_OPTIONS = ExpandOptions()


def except_json_ld_error(err: jsonld.JsonLdError):
    """Handle JsonLdError."""
    # We need to drill down; for instance, `to_rdf()` raises an error which
    # contains an actual error from `expand()` in its `.cause` field.
    err = err.cause or err

    if isinstance(err, JSONDecodeError):
        raise InvalidJSONLiteral() from err

    match err.code:
        case LoadingRemoteContextFailed.code:
            raise LoadingRemoteContextFailed(
                context=err.details['url'],
                reason=str(err.details['cause']),
            ) from err

        case _:
            raise PyLDError(
                message=str(err),
                code=err.code,
            ) from err


@contextlib.contextmanager
def except_json_ld_errors():   # noqa: WPS238, C901
    """Convert pyld errors to typed YAML-LD exceptions."""
    try:  # noqa: WPS225
        yield
    except TypeError as err:
        if 'not supported between instances of ' in str(err):
            raise MappingKeyError() from err

        raise
    except RecursionError as err:
        raise CycleDetected() from err
    except JSONDecodeError as err:
        raise InvalidScriptElement() from err
    except jsonld.JsonLdError as err:
        except_json_ld_error(err)


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def expand(   # noqa: C901, WPS211
    document: JsonLdInput,
    options: ExpandOptions = DEFAULT_EXPAND_OPTIONS,
) -> list[JsonLdRecord]:
    """
    Expand a [＊-LD](/blog/any-ld/) document.

    Converts all compact IRIs, keywords, and terms into their absolute IRI
    representations.
    """
    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)
    dict_options.setdefault('headers', {'Accept': DEFAULT_ACCEPT_HEADER})

    with except_json_ld_errors():
        jsonld._resolved_context_cache = jsonld.LRUCache(   # noqa: WPS437
            maxsize=jsonld.RESOLVED_CONTEXT_CACHE_MAX_SIZE,
        )
        return jsonld.expand(
            input_=ensure_string_or_document(document),
            options=dict_options,
        )



from pydantic import validate_call
from pyld import jsonld

from yaml_ld.document_loaders.default import DEFAULT_DOCUMENT_LOADER
from yaml_ld.expand import except_json_ld_errors
from yaml_ld.models import (
    DEFAULT_VALIDATE_CALL_CONFIG,
    JsonLdInput,
    JsonLdRecord,
    ensure_string_or_document,
)
from yaml_ld.options import (
    BaseOptions,
    ExpandContextOptions,
    ExtractAllScriptsOptions,
)


class FrameOptions(   # type: ignore
    BaseOptions,
    ExtractAllScriptsOptions,
    ExpandContextOptions,
):
    """Options for YAML-LD framing."""

    embed: bool = True
    """
    Sets the value object embed flag used in the Framing Algorithm. A boolean
    value of true sets the flag to @once, while a value of false sets the flag
    to @never.
    """

    explicit: bool = False
    """Default `@explicit` flag."""

    omit_default: bool = False
    """Default `@omit_default` flag."""

    prune_blank_node_identifiers: bool = True
    """Remove unnecessary blank node identifiers."""

    require_all: bool = False
    """Default `@require_all` flag."""

    frame_default: bool = False
    """Instead of framing a merged graph, frame only the default graph."""


DEFAULT_FRAME_OPTIONS = FrameOptions()


@validate_call(config=DEFAULT_VALIDATE_CALL_CONFIG)
def frame(
    document: JsonLdInput,
    frame: JsonLdRecord,   # noqa: WPS442
    options: FrameOptions = DEFAULT_FRAME_OPTIONS,
) -> JsonLdRecord:
    """Frame a [＊-LD](/blog/any-ld/) document."""
    dict_options = options.model_dump(by_alias=True, exclude_none=True)
    dict_options.setdefault('documentLoader', DEFAULT_DOCUMENT_LOADER)

    with except_json_ld_errors():
        return jsonld.frame(
            input_=ensure_string_or_document(document),
            frame=frame,
            options=dict_options,
        )
